/*
PMP Mod
Copyright (C) 2006 jonny

Homepage: http://jonny.leffe.dnsalias.com
E-mail:   jonny@leffe.dnsalias.com

This program is free software; you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation; either version 2 of the License, or
(at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
*/


#include <pspdebug.h>
#include <psppower.h>
#include <pspusb.h>
#include <pspusbstor.h>
#include "avformat.h"
#include "me.h"
#include "me_idct.h"
#include "pmp.h"
#include "opendir.h"

/* Get the full path to EBOOT.PBP. */
char psp_full_path[1024 + 1];



PSP_MODULE_INFO("pmpmod", 0x1000, 1, 1);
PSP_MAIN_THREAD_ATTR(0);


#define video_directory "ms0:/PSP/VIDEO/"


void pmp_pause()
	{
	while (1)
		{
		SceCtrlData controller;
		sceCtrlReadBufferPositive(&controller, 1);

		if (controller.Buttons & PSP_CTRL_CROSS || controller.Buttons & PSP_CTRL_HOME)
			{
			break;
			}
		}
	}


void screen_init()
	{
	pspDebugScreenSetTextColor(0xffffff);
	pspDebugScreenSetBackColor(0x000000);
	pspDebugScreenInit();
	}


void print_error(char *s)
	{
	screen_init();
	pspDebugScreenSetXY(0, 0);
	pspDebugScreenPrintf("Error:\n");
	pspDebugScreenPrintf("\n");
	pspDebugScreenPrintf("\"%s\"\n", s);
	pspDebugScreenPrintf("\n");
	pspDebugScreenPrintf("Press X");
	pmp_pause();
	}


char *static_init()
	{
	scePowerSetClockFrequency(333, 333, 166);
	scePowerLock(0);


	sceCtrlSetSamplingCycle(0);
	sceCtrlSetSamplingMode(PSP_CTRL_MODE_ANALOG);


	av_register_all();


	if (me_struct_init() == 0)
		{
		return("static_init: me_struct_init failed");
		}


	if (me_idct_p_malloc() == 0)
		{
		return("static_init: me_idct_p_malloc failed");
		}


	return(0);
	}


int init()
	{
	char *result = static_init();
	if (result != 0)
		{
        print_error(result);

		return(0);
		}


	return(1);
	}


void print_pmp(int licence)
	{
	screen_init();
	pspDebugScreenSetXY(0, 0);
	pspDebugScreenPrintf("PMP Mod v1.01 (M4) by jonny\n");
	pspDebugScreenPrintf("\n");
	pspDebugScreenPrintf("HOME = exit");
	
	
	if (licence == 0)
		{
		pspDebugScreenPrintf(", X = play, SQUARE = licence, START = refresh list");
		}


	pspDebugScreenSetXY(0, 23);
	pspDebugScreenPrintf("PMP Mod uses libavcodec from FFMPEG (http://ffmpeg.sourceforge.net/)\n");
	pspDebugScreenPrintf("Many thanks goes to:\n");
	pspDebugScreenPrintf("\n");
	pspDebugScreenPrintf("- FFMPEG developers\n");
	pspDebugScreenPrintf("- JiniCho (PMP Mod is a modified version of his original FFMPEG and\n");
	pspDebugScreenPrintf("  FFPLAY psp port)\n");
	pspDebugScreenPrintf("- ps2dev.org developers and users\n");
	pspDebugScreenPrintf("\n");
	pspDebugScreenPrintf("                                   http://jonny.leffe.dnsalias.com/");
	}


void print_licence()
	{
	pspDebugScreenPrintf("This program is free software; you can redistribute it and/or modify");
	pspDebugScreenPrintf("it under the terms of the GNU General Public License as published by");
	pspDebugScreenPrintf("the Free Software Foundation; either version 2 of the License, or\n");
	pspDebugScreenPrintf("(at your option) any later version.\n");
	pspDebugScreenPrintf("\n");
	pspDebugScreenPrintf("This program is distributed in the hope that it will be useful,\n");
	pspDebugScreenPrintf("but WITHOUT ANY WARRANTY; without even the implied warranty of\n");
	pspDebugScreenPrintf("MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the\n");
	pspDebugScreenPrintf("GNU General Public License for more details.\n");
	pspDebugScreenPrintf("\n");
	pspDebugScreenPrintf("You should have received a copy of the GNU General Public License\n");
	pspDebugScreenPrintf("along with this program; if not, write to the Free Software\n");
	pspDebugScreenPrintf("Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.\n");
	pspDebugScreenPrintf("\n");
	}


//------------------------------------

void initUSBdrivers(void)
{
int exec_pu;
   
   exec_pu=sceKernelLoadModule("flash0:/kd/semawm.prx",0,NULL);
   sceKernelStartModule(exec_pu, 0, NULL, 0, NULL);

   exec_pu=sceKernelLoadModule("flash0:/kd/usbstor.prx",0,NULL);
   sceKernelStartModule(exec_pu, 0, NULL, 0, NULL);

   exec_pu=sceKernelLoadModule("flash0:/kd/usbstormgr.prx",0,NULL);
   sceKernelStartModule(exec_pu, 0, NULL, 0, NULL);

   exec_pu=sceKernelLoadModule("flash0:/kd/usbstorms.prx",0,NULL);
   sceKernelStartModule(exec_pu, 0, NULL, 0, NULL);

   exec_pu=sceKernelLoadModule("flash0:/kd/usbstorboot.prx",0,NULL);
   sceKernelStartModule(exec_pu, 0, NULL, 0, NULL);

   int ret = 0;

// FIXME: last arg should be NULL (wrong prototype in pspsdk?)
   sceUsbStart("USBBusDriver",0,0);
   ret = sceUsbStart("USBStor_Driver",0,0);
   
   if(ret)
   {
// prototype in pspsdk has only one arg, is that correct?
//      sceUsbstorBootSetCapacity(0x00800000,ret);
      sceUsbstorBootSetCapacity(0x00800000);
   }

	sceUsbActivate(0x1c8);
} 

void exitUSBdrivers(void)
{
   sceUsbDeactivate();

// FIXME: last arg should be NULL (wrong prototype in pspsdk?)
   sceUsbStop("USBStor_Driver",0,0);
   sceUsbStop("USBBusDriver",0,0);
} 

int	strcmpupr( char* s1, char* s2 )
{
	while (*s1 && *s2)
	{
		char c1 = *s1++;
		char c2 = *s2++;

		if ((c1 >= 'a') && (c1 <= 'z'))
			c1 -= 'a' - 'A';
		if ((c2 >= 'a') && (c2 <= 'z'))
			c2 -= 'a' - 'A';

		if (c1 > c2)
			return 1;
		if (c1 < c2)
			return -1;
	}

	if (*s1)
		return 1;
	if (*s2)
		return -1;

	return 0;
}

//------------------------------

void main_loop()
	{


	while (1)
		{

		struct opendir_struct directory;

		char *result = opendir_open(&directory, video_directory);
		if (result != 0)
			{
			print_pmp(1);
			pspDebugScreenSetXY(0, 4);
			pspDebugScreenPrintf("\"%s\" not found or empty", video_directory);
			pspDebugScreenSetXY(0, 8);
			print_licence();
			pmp_pause();
			return;
			}


		//--------------------------
		// sort files
		int		number_of_files = 0;
		char**	sorted_files = NULL;
		{
			int		i;

			for (i = 0; i < directory.number_of_directory_entries; i++)
			{
				char*	name = directory.directory_entry[i].d_name;

				if ((strcmp( name + strlen(name) - 4, ".PMP") == 0) ||
					(strcmp( name + strlen(name) - 4, ".pmp") == 0))
					number_of_files++;
			}

			if (number_of_files)
			{
				sorted_files = (char**)malloc_64( sizeof(char*) * number_of_files );

				int	n = 0;

				for (i = 0; i < directory.number_of_directory_entries; i++)
				{
					char	name[1024];
					strcpy( name, directory.directory_entry[i].d_name );
					int		len = strlen( name );

					if ((strcmp( name + len - 4, ".PMP") == 0) ||
						(strcmp( name + len - 4, ".pmp") == 0))
					{
						// save pointer to entry name
						sorted_files[n++] = directory.directory_entry[i].d_name;
					}
				}

				// bubble sort
				int		swap = 1;
				while (swap)
				{
					swap = 0;

					for (i = 0; i < number_of_files - 1; i++)
					{
						if (strcmpupr( sorted_files[i] , sorted_files[i+1]) > 0)
						{
							swap = 1;
							char*	name = sorted_files[i];
							sorted_files[i] = sorted_files[i+1];
							sorted_files[i+1] = name;
						}
					}
				}
			}		
		}

		//--------------------------



		int selected_entry         = 0;
		int top_entry              = 0;
		int maximum_number_of_rows = 16;
		int starting_row           = 4;


		print_pmp(0);


		//----------------
		initUSBdrivers();
		//----------------


        SceCtrlData controller;

		while (1)
			{
			sceCtrlReadBufferPositive(&controller, 1);

			if (controller.Buttons & PSP_CTRL_CROSS || controller.Buttons & PSP_CTRL_HOME || controller.Buttons & PSP_CTRL_SQUARE  || controller.Buttons & PSP_CTRL_START)
				{
				break;
				}


			if (controller.Buttons & PSP_CTRL_DOWN)
				{
				if (selected_entry + 1 < number_of_files)
					{
					selected_entry++;

					if (selected_entry == top_entry + maximum_number_of_rows)
						{
						top_entry++;
						}
					}
				}


			if (controller.Buttons & PSP_CTRL_UP)
				{
				if (selected_entry != 0)
					{
					selected_entry--;

					if (selected_entry == top_entry - 1)
						{
						top_entry--;
						}
					}
				}




			pspDebugScreenSetXY(1, starting_row);
			pspDebugScreenSetTextColor(0xffffff);
			pspDebugScreenSetBackColor(0xaa4400);
			pspDebugScreenPrintf("%-66.66s", "");


			int i = 0;
			for (; i < maximum_number_of_rows; i++)
				{
				int current_entry = top_entry + i;


				pspDebugScreenSetXY(1, starting_row + i + 1);

				if (current_entry < number_of_files)
					{
					if (current_entry == selected_entry)
						{
						pspDebugScreenSetTextColor(0xffffff);
						pspDebugScreenSetBackColor(0x882200);
						}
					else
						{
						pspDebugScreenSetTextColor(0xffffff);
						pspDebugScreenSetBackColor(0xcc6600);
						}

					pspDebugScreenPrintf("%-66.66s", sorted_files[current_entry]);
					}
				else
					{
					pspDebugScreenSetTextColor(0xffffff);
					pspDebugScreenSetBackColor(0xaa4400);
					pspDebugScreenPrintf("%-66.66s", "");
					}
				}


			pspDebugScreenSetXY(1, starting_row + maximum_number_of_rows + 1);
			pspDebugScreenSetTextColor(0xffffff);
			pspDebugScreenSetBackColor(0xaa4400);
			pspDebugScreenPrintf("%-66.66s", "");


			sceKernelDelayThread(100000);
			}

		//-----------------
		exitUSBdrivers();
		//-----------------



		if (controller.Buttons & PSP_CTRL_HOME)
			{
			break;
			}


		if (controller.Buttons & PSP_CTRL_SQUARE)
			{
			screen_init();
			pspDebugScreenSetXY(0, 0);
			print_licence();
			pspDebugScreenPrintf("Press X");
			pmp_pause();
			screen_init();
			sceKernelDelayThread(1000000);
			continue;
			}

		// just refresh
		if (controller.Buttons & PSP_CTRL_START)	
			continue;


		screen_init();
		pspDebugScreenSetXY(0, 0);
		pspDebugScreenPrintf("Loading ...");

		sceKernelDelayThread(100000);


		result = pmp_play(sorted_files[selected_entry]);
		if (result != 0)
			{
			sceKernelDelayThread(1000000);
			print_error(result);
			screen_init();
			sceKernelDelayThread(1000000);
			}


		//-----------------
		free_64(sorted_files);
		//-----------------

		opendir_close(&directory);


		}


	}


int main(int argc, char *argv[])
	{
	char *psp_eboot_path;
 
	strncpy(psp_full_path, argv[0], sizeof(psp_full_path) - 1);
	psp_full_path[sizeof(psp_full_path) - 1] = '\0';
 
	psp_eboot_path = strrchr(psp_full_path, '/');
	if (psp_eboot_path != NULL)
	    {
		*psp_eboot_path = '\0';
		} 


	if (init())
		{
		main_loop();
		}


	sceKernelExitGame();
	return(0);
	}
